#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Python Motif Enrichment Analysis Library
# Bjørn Bredesen, 2018
# bjorn.bredesen@ii.uib.no
############################################################################
# Python library for analysing and modelling motif occurrence in DNA
# sequences.
############################################################################
# Interfacing with scikit-learn

from .models import sequenceModel
import numpy as np
from sklearn import svm
from sklearn.ensemble import RandomForestClassifier
import random

# Support Vector Machines
class sequenceModelSVM(sequenceModel):
	def __init__(self, features, positives, negatives, windowSize, windowStep, kDegree):
		self.features = features
		self.windowSize, self.windowStep = windowSize, windowStep
		self.positives, self.negatives = positives, negatives
		self.kernel = kDegree
		self.threshold = 0.0
		vP = [ self.getSequenceFeatureVector(w) for s in positives for w in s.getWindows(self.windowSize, self.windowStep) ]
		vN = [ self.getSequenceFeatureVector(w) for s in negatives for w in s.getWindows(self.windowSize, self.windowStep) ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		self.cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		self.cls.fit( np.array(vP+vN), np.array(cP+cN) )
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	def scoreWindow(self, seq):
		return self.cls.decision_function(np.array([self.getSequenceFeatureVector(seq)]))[0]
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Positives: %s; Negatives: %s; Kernel: %s; Support vectors: %d>'%(str(self.features), len(self.features), str(self.positives), str(self.negatives), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], len(self.cls.support_vectors_))
	def __repr__(self): return self.__str__()

# Support Vector Machines
class sequenceModelSVMOptimizedQuadratic(sequenceModel):
	def __init__(self, features, positives, negatives, windowSize, windowStep, kDegree):
		self.features = features
		self.windowSize, self.windowStep = windowSize, windowStep
		self.positives, self.negatives = positives, negatives
		self.kernel = kDegree
		self.threshold = 0.0
		assert(kDegree == 2)
		vP = [ self.getSequenceFeatureVector(w) for s in positives for w in s.getWindows(self.windowSize, self.windowStep) ]
		vN = [ self.getSequenceFeatureVector(w) for s in negatives for w in s.getWindows(self.windowSize, self.windowStep) ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		cls.fit( np.array(vP+vN), np.array(cP+cN) )
		assert(cls.coef0 == 0.0)
		# Extract basic kernel values
		SV = cls.support_vectors_
		nfeatures = len(cls.support_vectors_[0])
		bias = cls.intercept_[0]
		c0 = cls.coef0
		c = cls.dual_coef_[0]
		gamma = 1.0 / (nfeatures)
		self.nSV = len(cls.support_vectors_)
		# Extract weight matrix
		"""
		self.pairWeights = np.zeros((nfeatures, nfeatures))
		for i in range(nfeatures):
			for j in range(nfeatures):
				self.pairWeights[i][j] = gamma*gamma*sum(c[vi]*v[i]*v[j] for vi, v in enumerate(SV))
		"""
		self.pairWeights = ((SV.T * c) @ SV) * gamma * gamma
		# Extract Bias
		self.bias = -bias
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	def scoreWindow(self, seq):
		fv = np.array(self.getSequenceFeatureVector(seq))
		return self.pairWeights @ fv @ fv - self.bias
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Positives: %s; Negatives: %s; Kernel: %s; Support vectors: %d; Quadratic optimized>'%(str(self.features), len(self.features), str(self.positives), str(self.negatives), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], self.nSV)
	def __repr__(self): return self.__str__()

#----------------------------------------------------------------

# Support Vector Machines
class sequenceModelEnsembleSVM(sequenceModel):
	def __init__(self, features, positives, negatives, windowSize, windowStep, kDegree, nEnsemble):
		self.features = features
		self.windowSize, self.windowStep = windowSize, windowStep
		self.positives, self.negatives = positives, negatives
		self.kernel = kDegree
		self.threshold = 0.0
		self.ensemble = []
		for _ in range(nEnsemble):
			epos = [ random.choice(positives) for _ in range(len(positives)) ]
			eneg = [ random.choice(negatives) for _ in range(len(negatives)) ]
			vP = [ self.getSequenceFeatureVector(w) for s in epos for w in s.getWindows(self.windowSize, self.windowStep) ]
			vN = [ self.getSequenceFeatureVector(w) for s in eneg for w in s.getWindows(self.windowSize, self.windowStep) ]
			cP = [ 1.0 for _ in range(len(vP)) ]
			cN = [ -1.0 for _ in range(len(vN)) ]
			cls = svm.SVC(kernel = 'poly', degree = kDegree)
			cls.fit( np.array(vP+vN), np.array(cP+cN) )
			self.ensemble.append(cls)
	def getCrossValidationConstructor(self):
		return lambda pos, neg: sequenceModelEnsembleSVM(self.features, pos, neg, self.windowSize, self.windowStep, self.kDegree, self.nEnsemble, self.nTrain)
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
		#return [ f.get(seq) for f in self.features ]
	def scoreWindow(self, seq):
		f = np.array([self.getSequenceFeatureVector(seq)])
		return sum(cls.decision_function(f)[0] for cls in self.ensemble) / len(self.ensemble)
		#return self.cls.decision_function(f)[0]
	def __str__(self):
		return 'Support Vector Machine Ensemble<Classifiers: %d; Features: %s; Positives: %s; Negatives: %s; Kernel: %s>'%(len(self.ensemble), str(self.features), str(self.positives), str(self.negatives), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1])
	def __repr__(self): return self.__str__()

# Random Forest
class sequenceModelRF(sequenceModel):
	def __init__(self, features, positives, negatives, windowSize, windowStep, nTrees, maxDepth):
		self.features = features
		self.windowSize, self.windowStep = windowSize, windowStep
		self.positives, self.negatives = positives, negatives
		self.nTrees, self.maxDepth = nTrees, maxDepth
		self.threshold = 0.0
		vP = [ self.getSequenceFeatureVector(w) for s in positives for w in s.getWindows(self.windowSize, self.windowStep) ]
		vN = [ self.getSequenceFeatureVector(w) for s in negatives for w in s.getWindows(self.windowSize, self.windowStep) ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		self.cls = RandomForestClassifier(max_depth = maxDepth, random_state = 0)
		self.cls.fit( np.array(vP+vN), np.array(cP+cN) )
	def getCrossValidationConstructor(self):
		return lambda pos, neg: sequenceModelRF(self.features, pos, neg, self.windowSize, self.windowStep, self.nTrees, self.maxDepth)
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
		#return [ f.get(seq) for f in self.features ]
	def scoreWindow(self, seq):
		return float(self.cls.predict_proba(np.array([self.getSequenceFeatureVector(seq)]))[0][1])
		#self.cls.decision_function(np.array([self.getSequenceFeatureVector(seq)]))[0]
	def __str__(self):
		return 'Random Forest<Features: %s; Positives: %s; Negatives: %s; Trees: %d; Max. depth: %d>'%(str(self.features), str(self.positives), str(self.negatives), self.nTrees, self.maxDepth)
	def __repr__(self): return self.__str__()


