#!/usr/bin/python3
# -*- coding: latin-1 -*-
############################################################################
# Python Motif Enrichment Analysis Library
# Bjørn Bredesen, 2018
# bjorn.bredesen@ii.uib.no
############################################################################
# Python library for analysing and modelling motif occurrence in DNA
# sequences.
############################################################################
# Interfacing with scikit-learn

from .models import sequenceModel
import numpy as np
import cupy as cp
from sklearn import svm
from sklearn.ensemble import RandomForestClassifier
import random

# Support Vector Machines
class sequenceModelSVM(sequenceModel):
	def __init__(self, features, positives, negatives, windowSize, windowStep, kDegree):
		self.features = features
		self.windowSize, self.windowStep = windowSize, windowStep
		self.positives, self.negatives = positives, negatives
		self.kernel = kDegree
		self.threshold = 0.0
		vP = [ self.getSequenceFeatureVector(w) for s in positives for w in s.getWindows(self.windowSize, self.windowStep) ]
		vN = [ self.getSequenceFeatureVector(w) for s in negatives for w in s.getWindows(self.windowSize, self.windowStep) ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		self.cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		self.cls.fit( np.array(vP+vN), np.array(cP+cN) )
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	def scoreWindow(self, seq):
		return self.cls.decision_function(np.array([self.getSequenceFeatureVector(seq)]))[0]
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Positives: %s; Negatives: %s; Kernel: %s; Support vectors: %d>'%(str(self.features), len(self.features), str(self.positives), str(self.negatives), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], len(self.cls.support_vectors_))
	def __repr__(self): return self.__str__()

# Support Vector Machines
class sequenceModelSVMOptimizedQuadratic(sequenceModel):
	def __init__(self, features, positives, negatives, windowSize, windowStep, kDegree, manualSummation = False):
		self.features = features
		self.windowSize, self.windowStep = windowSize, windowStep
		self.positives, self.negatives = positives, negatives
		self.kernel = kDegree
		self.threshold = 0.0
		self.manualSummation = manualSummation
		assert(kDegree == 2)
		vP = [ self.getSequenceFeatureVector(w) for s in positives for w in s.getWindows(self.windowSize, self.windowStep) ]
		vN = [ self.getSequenceFeatureVector(w) for s in negatives for w in s.getWindows(self.windowSize, self.windowStep) ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		cls.fit( np.array(vP+vN), np.array(cP+cN) )
		assert(cls.coef0 == 0.0)
		# Extract basic kernel values
		SV = cls.support_vectors_
		nfeatures = len(cls.support_vectors_[0])
		bias = cls.intercept_[0]
		c0 = cls.coef0
		c = cls.dual_coef_[0]
		gamma = 1.0 / (nfeatures)
		self.nSV = len(cls.support_vectors_)
		# Extract weight matrix
		self.pairWeights = ((SV.T * c) @ SV) * gamma * gamma
		# Extract Bias
		self.bias = -bias
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	def scoreWindow(self, seq):
		if self.manualSummation:
			fv = np.array(self.getSequenceFeatureVector(seq))
			nf = len(fv)
			return sum( self.pairWeights[ x, y ] * fv[x] * fv[y] for x in range(nf) for y in range(nf) ) - self.bias
			#return self.pairWeights @ fv @ fv - self.bias
		else:
			fv = np.array(self.getSequenceFeatureVector(seq))
			return self.pairWeights @ fv @ fv - self.bias
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Positives: %s; Negatives: %s; Kernel: %s; Support vectors: %d; Quadratic optimized>'%(str(self.features), len(self.features), str(self.positives), str(self.negatives), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], self.nSV)
	def __repr__(self): return self.__str__()


# Support Vector Machines
class sequenceModelSVMOptimizedQuadraticCUDA(sequenceModel):
	def __init__(self, features, positives, negatives, windowSize, windowStep, kDegree, useBaseScoring = False):
		self.features = features
		self.windowSize, self.windowStep = windowSize, windowStep
		self.positives, self.negatives = positives, negatives
		self.kernel = kDegree
		self.threshold = 0.0
		self.useBaseScoring = useBaseScoring
		assert(kDegree == 2)
		vP = [ self.getSequenceFeatureVector(w) for s in positives for w in s.getWindows(self.windowSize, self.windowStep) ]
		vN = [ self.getSequenceFeatureVector(w) for s in negatives for w in s.getWindows(self.windowSize, self.windowStep) ]
		cP = [ 1.0 for _ in range(len(vP)) ]
		cN = [ -1.0 for _ in range(len(vN)) ]
		cls = svm.SVC(kernel = 'poly', degree = kDegree, gamma = 'auto')
		cls.fit( np.array(vP+vN), np.array(cP+cN) )
		assert(cls.coef0 == 0.0)
		# Extract basic kernel values
		SV = cls.support_vectors_
		nfeatures = len(cls.support_vectors_[0])
		bias = cls.intercept_[0]
		c0 = cls.coef0
		c = cls.dual_coef_[0]
		gamma = 1.0 / (nfeatures)
		self.nSV = len(cls.support_vectors_)
		# Extract weight matrix
		self.pairWeights = cp.array( ((SV.T * c) @ SV) * gamma * gamma )
		# Extract Bias
		self.bias = -bias
	def getSequenceFeatureVector(self, seq):
		return self.features.getAll(seq)
	def scoreWindow(self, seq):
		fv = cp.array(self.getSequenceFeatureVector(seq))
		return self.pairWeights @ fv @ fv - self.bias
	def __str__(self):
		return 'Support Vector Machine<Features: %s (%d); Positives: %s; Negatives: %s; Kernel: %s; Support vectors: %d; Quadratic optimized, CUDA>'%(str(self.features), len(self.features), str(self.positives), str(self.negatives), [ 'linear', 'quadratic', 'cubic' ][self.kernel-1], self.nSV)
	def __repr__(self): return self.__str__()

